(module client mzscheme
  
  ;; ---------------------------------------- interface ----------------------------------------
  
  ;; player-packages : num (listof player) (listof package) -> (listof package)
  
  ;; run-client : (board num num -> void) (num (listof player) (listof package) (listof package) (listof update) num package-data (listof posn) player -> command) -> void
  ;; where package-data = (listof (U (list num num) (list num num num)))
  
  ;; the first arg is a function that consumes the initialization data:
  ;; - the board
  ;; - your id
  ;; - weight capacity
  ;; and doesn't produce a value
  
  ;; the second arg is a function that consumes the state at each round:
  ;; - money left
  ;; - players
  ;; - known packages
  ;; - here packages
  ;; - last updates
  ;; - your current score
  ;; - package data ...
  ;; - possible package locations
  ;; - me
  ;; and produces a command
  
  ;; -------------------------------------------------------------------------------------------
  
  (require (lib "list.ss")
           (lib "match.ss")
           (lib "posn.ss" "lang")
           "board.ss"
           "types.ss"
           "utils.ss")
  
  (provide run-client
           player-packages)
  
  ;; read-config : input-port -> (values num num num)
  (define (read-config in)
    (match-let ([(id capacity money) (map string->number (read-list in))])
      (values id capacity money)))
  
  (define (de-hash s)
    (substring s 1 (string-length s)))
  
  ;; read-init-players : input-port -> (listof player)
  (define (read-init-players in)
    (let loop ([l (read-list in)])
      (match l
        [() '()]
        [(id "X" x "Y" y . r) (cons (make-player (string->number (de-hash id))
                                                 (string->number x)
                                                 (string->number y)
                                                 '())
                                    (loop r))])))
  
  ;; read-packages : input-port -> (listof package)
  (define (read-packages in)
    (let loop ([l (read-list in)])
      (match l
        [() '()]
        [(id dest-x dest-y weight . r) (cons (make-package (string->number id)
                                                           (string->number dest-x)
                                                           (string->number dest-y)
                                                           (string->number weight))
                                             (loop r))])))
  
  (define (dir->char dir)
    (case dir
      [(N) #\N]
      [(S) #\S]
      [(E) #\E]
      [(W) #\W]))
  
  ;; send-command : output-port command -> void
  (define (send-command out cmd)
    (let ([s (string-append (number->string (command-bid cmd))
                            " "
                            (match (command-action cmd)
                              [($ move dir) (format "Move ~a" (dir->char dir))]
                              [($ pick ids) (format "Pick~a" (foldr (lambda (id s) (string-append " " (number->string id) s)) "" ids))]
                              [($ drop ids) (format "Drop~a" (foldr (lambda (id s) (string-append " " (number->string id) s)) "" ids))])
                            "\n")])
      (printf "SENT: ~a" s)
      (fprintf out s)))
  
  ;  (define (test4)
  ;    (send-command (current-output-port) (make-command 200 (make-pick (list 1 2 3 4 5)))))
  
  (define (starts-hash? s)
    (equal? #\# (string-ref s 0)))
  
  (define (group-by-hash l)
    (let-values ([(_ all) (let loop ([m l])
                            (if (null? m)
                                (values '() '())
                                (let-values ([(curr all) (loop (cdr m))])
                                  (if (starts-hash? (car m))
                                      (values '() (cons (cons (de-hash (car m)) curr) all))
                                      (values (cons (car m) curr) all)))))])
      all))
  
  ;; read-reply : input-port -> (listof update)
  (define (read-reply in)
    (let loop-player ([ps (group-by-hash (read-list in))])
      (if (null? ps)
          '()
          (let ([p (car ps)])
            (cons (make-update (string->number (car p))
                               (let loop ([a (cdr p)])
                                 (match a
                                   [() '()]
                                   [("N" . r) (cons (make-u-move 'N) (loop r))]
                                   [("S" . r) (cons (make-u-move 'S) (loop r))]
                                   [("E" . r) (cons (make-u-move 'E) (loop r))]
                                   [("W" . r) (cons (make-u-move 'W) (loop r))]
                                   [("P" id . r) (cons (make-u-pick (string->number id)) (loop r))]
                                   [("D" id . r) (cons (make-u-drop (string->number id)) (loop r))]
                                   [("X" x "Y" y . r) (cons (make-u-loc (string->number x) (string->number y)) (loop r))])))
                  (loop-player (cdr ps)))))))
  
  ;  (define (test2)
  ;    (read-reply (open-input-string "#2 E #1 N #3 W #4 P 0 P 1 #5 E #6 W E #7 X 1 Y 1 #8 E")))
  
  ;; update-player : u-action player -> player
  (define (update-player a p)
    (match a
      [($ u-move 'N) (make-player (player-id p) (player-x p) (+ (player-y p) 1) (player-package-ids p))]
      [($ u-move 'S) (make-player (player-id p) (player-x p) (- (player-y p) 1) (player-package-ids p))]
      [($ u-move 'E) (make-player (player-id p) (+ (player-x p) 1) (player-y p) (player-package-ids p))]
      [($ u-move 'W) (make-player (player-id p) (- (player-x p) 1) (player-y p) (player-package-ids p))]
      [($ u-pick id) (make-player (player-id p) (player-x p) (player-y p) (cons id (player-package-ids p)))]
      [($ u-drop id) (make-player (player-id p) (player-x p) (player-y p) (remove id (player-package-ids p)))]
      [($ u-loc x y) (make-player (player-id p) x y (player-package-ids p))]))
  
  ;; apply-update : (listof r-action) player -> player
  (define (apply-update as p)
    (if (null? as)
        p
        (apply-update (cdr as) (update-player (car as) p))))
  
  ;; update-players : (listof player) (listof update) -> (listof player)
  (define (update-players players updates)
    (let loop ([ps (quicksort players (lambda (a b) (< (player-id a) (player-id b))))]
               [us (quicksort updates (lambda (a b) (< (update-id a) (update-id b))))])
      (if (null? us)
          '() ;; remaining players are dead
          (let ([u (car us)])
            (cond
              [(or (null? ps)
                   (> (player-id (car ps)) (update-id u))) (cons (apply-update (update-u-actions u) (make-player (update-id u) 'unknown-x 'unknown-y '()))
                                                                 (loop ps (cdr us)))] ;; player appeared
              [(< (player-id (car ps)) (update-id u)) (loop (cdr ps) us)] ;; player died
              [else (cons (apply-update (update-u-actions u) (car ps))
                          (loop (cdr ps) (cdr us)))])))))
  
  ;; sort-packages : (listof package) -> (listof package)
  (define (sort-packages l)
    (quicksort l (lambda (a b) (< (package-id a) (package-id b)))))
  
  ;; merge-packages : (listof package) (listof package) -> (listof package)
  (define (merge-packages l m)
    (cond
      [(null? l) m]
      [(null? m) l]
      [else (let ([lid (package-id (car l))]
                  [mid (package-id (car m))])
              (cond
                [(< lid mid) (cons (car l) (merge-packages (cdr l) m))]
                [(> lid mid) (cons (car m) (merge-packages l (cdr m)))]
                [else (cons (car l) (merge-packages (cdr l) (cdr m)))]))]))
  
  ;; lookup-update : num (listof update) -> (U #f update)
  (define (lookup-update id us)
    (cond
      [(null? us) #f]
      [(equal? id (update-id (car us))) (car us)]
      [else (lookup-update id (cdr us))]))
  
  ;; lookup-player : num (listof player) -> (U #f player)
  (define (lookup-player id ps)
    (cond
      [(null? ps) #f]
      [(equal? id (player-id (car ps))) (car ps)]
      [else (lookup-player id (cdr ps))]))
  
  ;; lookup-package : num (listof packages) -> (U #f package)
  (define (lookup-package id ps)
    (cond
      [(null? ps) #f]
      [(equal? id (package-id (car ps))) (car ps)]
      [else (lookup-package id (cdr ps))]))
  
  ;; new-score : num (listof update) (listof player) (listof package) -> num
  (define (new-score my-id updates players packages)
    (let ([my-update (lookup-update my-id updates)])
      (if (equal? my-update #f)
          0
          (let ([me (lookup-player my-id players)])
            (if (equal? me #f)
                (error 'new-score "unknown player, id ~s" my-id)
                (let loop ([score 0]
                           [as (update-u-actions my-update)]
                           [p me])
                  (if (null? as)
                      score
                      (match (car as)
                        [($ u-move 'N) (loop score (cdr as) (make-player (player-id p) (player-x p) (+ (player-y p) 1) (player-package-ids p)))]
                        [($ u-move 'S) (loop score (cdr as) (make-player (player-id p) (player-x p) (- (player-y p) 1) (player-package-ids p)))]
                        [($ u-move 'E) (loop score (cdr as) (make-player (player-id p) (+ (player-x p) 1) (player-y p) (player-package-ids p)))]
                        [($ u-move 'W) (loop score (cdr as) (make-player (player-id p) (- (player-x p) 1) (player-y p) (player-package-ids p)))]
                        [($ u-pick id) (loop score (cdr as) (make-player (player-id p) (player-x p) (player-y p) (cons id (player-package-ids p))))]
                        [($ u-drop id) (loop (+ score (let ([pkg (lookup-package id packages)])
                                                        (if (equal? pkg #f)
                                                            (error 'new-score "dropped unknown package, id ~s" id)
                                                            (if (and (equal? (package-dest-x pkg) (player-x p))
                                                                     (equal? (package-dest-y pkg) (player-y p)))
                                                                (package-weight pkg)
                                                                0))))
                                             (cdr as)
                                             (make-player (player-id p) (player-x p) (player-y p) (remove id (player-package-ids p))))]
                        [($ u-loc x y) (loop score (cdr as) (make-player (player-id p) x y (player-package-ids p)))]))))))))
  
  ;; update-package-data : (listof (U (list num num) (list num num num))) (listof update) (listof player) -> (listof (U (list num num) (list num num num)))
  (define (update-package-data package-data updates players)
    (define (drop-package id x y pd)
      (if (null? pd)
          (list (list id x y)) ;; can this ever happen?
          (if (equal? id (caar pd))
              (cons (list id x y) (cdr pd))
              (cons (car pd) (drop-package id x y (cdr pd))))))
    (if (null? updates)
        package-data
        (let* ([u (car updates)]
               [plid (update-id u)])
          (let loop ([as (update-u-actions u)]
                     [pd package-data]
                     [p (lookup-player plid players)])
            (if (or (null? as) (equal? p #f))
                (update-package-data pd (cdr updates) players)
                (match (car as)
                  [($ u-move 'N) (loop (cdr as) pd (make-player (player-id p) (player-x p) (+ (player-y p) 1) (player-package-ids p)))]
                  [($ u-move 'S) (loop (cdr as) pd (make-player (player-id p) (player-x p) (- (player-y p) 1) (player-package-ids p)))]
                  [($ u-move 'E) (loop (cdr as) pd (make-player (player-id p) (+ (player-x p) 1) (player-y p) (player-package-ids p)))]
                  [($ u-move 'W) (loop (cdr as) pd (make-player (player-id p) (- (player-x p) 1) (player-y p) (player-package-ids p)))]
                  [($ u-pick id) (loop (cdr as) (cons (list id plid) pd) (make-player (player-id p) (player-x p) (player-y p) (cons id (player-package-ids p))))]
                  [($ u-drop id) (loop (cdr as) (drop-package id (player-x p) (player-y p) pd)
                                       (make-player (player-id p) (player-x p) (player-y p) (remove id (player-package-ids p))))]
                  [($ u-loc x y) (loop (cdr as) pd (make-player (player-id p) x y (player-package-ids p)))]))))))
  
  ;; all-homes : board -> (listof posn)
  (define (all-homes b)
    (let ([w (board-width b)]
          [h (board-height b)])
      (let loop ([x 1] [y 1])
        (cond
          [(> y h) '()]
          [(> x w) (loop 1 (add1 y))]
          [(equal? (board-ref b x y) 'home) (cons (make-posn x y) (loop (add1 x) y))]
          [else (loop (add1 x) y)]))))
  
  ;; update-homes : player (listof posn) -> (listof posn)
  (define (update-homes me homes)
    (let ([x (player-x me)]
          [y (player-y me)])
      (let loop ([h homes])
        (if (null? h)
            '()
            (if (and (equal? x (posn-x (car h)))
                     (equal? y (posn-y (car h))))
                (cdr h)
                (cons (car h) (loop (cdr h))))))))  
  
  ;; package-posns : package-data -> (listof posn)
  (define (package-posns l)
    (match l
      [() '()]
      [((_ _) . r) (package-posns r)]
      [((_ x y) . r) (cons (make-posn x y) (package-posns r))]))
  
  (define delete-duplicates
    (lambda (s)
      (if (null? s) s
          (let ((a (car s)) (d (cdr s)))
            (if (member a d) (delete-duplicates d)
                (cons a (delete-duplicates d)))))))
  
  ;; merge-posns : (listof posn) (listof posn) -> (listof posn)
  (define (merge-posns l1 l2)
    (delete-duplicates (append l1 l2)))
  
  ;; run-client : (board num num -> void) (num (listof player) (listof package) (listof package) (listof update) num package-data (listof posn) player -> command) -> void
  ;; where package-data = (listof (U (list num num) (list num num num)))
  (define (run-client init-handler round-handler)
    (with-handlers ([(lambda (exn) (equal? exn 'no-more-input)) (lambda (exn)
                                                                  (display "no more input ... bye")
                                                                  (newline))])
      (let ([argv (current-command-line-arguments)])
        (let* ([server-name (vector-ref argv 0)]
               [server-port (string->number (vector-ref argv 1))])
          (let-values ([(in out) (tcp-connect server-name server-port)])
            (fprintf out "Player\n")
            (let*-values ([(board) (read-board in)]
                          [(my-id capacity init-money) (read-config in)]
                          [(init-players) (read-init-players in)])
              (init-handler board my-id capacity)
              (let loop ([money-left init-money]
                         [players init-players]
                         [known-packages '()]
                         [updates '()]
                         [my-score 0]
                         [package-data '()]
                         [unvisited-homes (all-homes board)])
                (let ([here-packages (read-packages in)])
                  (let ([cmd (round-handler money-left players known-packages here-packages updates my-score package-data
                                            (merge-posns (package-posns package-data) unvisited-homes)
                                            (lookup-player my-id players))])
                    (send-command out cmd)
                    (let ([reply (read-reply in)])
                      (loop (- money-left (abs (command-bid cmd)))
                            (update-players players reply)
                            (merge-packages known-packages (sort-packages here-packages))
                            reply
                            (+ my-score (new-score my-id reply players known-packages))
                            (update-package-data package-data reply players)
                            (update-homes (lookup-player my-id players) unvisited-homes))))))))))))
  
  ;; player-packages : num (listof player) (listof package) -> (listof package)
  (define (player-packages pl-id players packages)
    (let ([me (lookup-player pl-id players)])
      (map (lambda (pkg-id) (lookup-package pkg-id packages)) (player-package-ids me))))
  
  )
