(module physics mzscheme
  (require "base-gm.ss"
	   "../dbt/board.ss"
	   "../dbt/types.ss"
	   "../rob/draw-util.scm"
	   "union.ss")

  (provide physics-apply-move! 
	   physics-apply-move-sequence 
	   cluster
	   my-cluster
	   move-player
	   player-possible-moves
	   player-tostring
	   player->hash
	   players->hash-pos
	   (all-from "../dbt/board.ss")
	   (all-from "../dbt/types.ss"))

  (define (player-tostring p)
    (if (not (player? p)) false
	(format "<pl: ~a ~a ~a>"
		(player-id p)
		(player-x p)
		(player-y p))))

  (define (player->hash p) (vector (player-id p) (player-x p) (player-y p)))
  (define (players->hash-pos players) (quicksort-vector! (list->vector (map player->hash players))
						    (lambda (a b) (< (vector-ref a 0)
								     (vector-ref b 0)))))
    
  (define debug-f (make-debug player-tostring))

  (define (move-player player dir)
    (make-player (player-id player)
		 (next-x (player-x player) dir)
		 (next-y (player-y player) dir)
		 (player-package-ids player)))

  (define (move-player! player dir)
    (set-player-x! player (next-x (player-x player) dir))
    (set-player-y! player (next-y (player-y player) dir)))


  (define (next-x x dir) 
    (case dir
      ('none x)
      ('N x)
      ('S x)
      ('E (+ x 1))
      ('W (- x 1))))

  (define (next-y y dir) 
    (case dir
      ('none y)
      ('N (+ y 1))
      ('S (- y 1))
      ('E y) 
      ('W y)))

  (define (here? p x y)
    (and (= (player-x p) x)
	 (= (player-y p) y)))

  (define (dead? p) (= (player-x p) -1))
  (define (kill! p) (set-player-x! p -1))

  (define (going-to-water? board x y dir)
    (eq? (board-ref board (next-x x dir) (next-y y dir)) 'water))

  (define (physics-apply-move! board players p dir)
    (let ((bumped (physics-apply-nudge! board players 
					(next-x (player-x p) dir)
					(next-y (player-y p) dir)
					dir)))
      (cond ((not bumped) false)
	    ((going-to-water? board (player-x p) (player-y p) dir) (kill! p) empty)
	    (true (move-player! p dir) bumped))))

  ;; returns false if the move hits a wall, otherwise return a list
  ;; of rebooted players. The players are modified destructively
  (define (physics-apply-nudge! board players x y dir)
    (cond ((eq? (board-ref board x y) 'wall) false)
	  ((not (ormap (lambda (p) (here? p x y)) players)) empty)
	  (true 
	   (let ((bumped (physics-apply-nudge! board players 
					       (next-x x dir) (next-y y dir)
					       dir)))
	     (if (not bumped) false
		 (let ((going-to-water (going-to-water? board x y dir)))
		   (for-each (lambda (p) (when (here? p x y)
					   (if going-to-water (kill! p)
					       (begin (move-player! p dir)
						      (set! bumped (cons p bumped))))))
			     players)
		   bumped))))))




  (define (physics-apply-move-sequence board players their-moves) 
    (let ((bumped (make-hash))
	  (copy (map (lambda (p) (make-player (player-id p)
					      (player-x p)
					      (player-y p)
					      (player-package-ids p))) 
		     players)))
      (let loop ((pls copy) (mvs their-moves))
	(if (empty? pls) (filter (lambda (p) (not (dead? p))) copy)
	    (begin (if (and (not (hash-get bumped (first pls) (lambda () false)))
			    (not (eq? (first mvs) 'none))
			    (not (dead? (first pls))))
		       (let ((bumped-now (physics-apply-move! board copy (first pls) (first mvs))))
			 (if bumped-now
			     (for-each (lambda (bd) (hash-put! bumped bd true))
				       bumped-now))))
		   (loop (rest pls) (rest mvs)))))))





  (define (cluster board players)
    (let* ((players (map make-union players))
	   (hash (make-hash 'equal)) ; (x . y) -> player-union
	   (apply (lambda (p x y)
		    (let ((tile (board-ref board x y)))
		      (when (is-occupable-tile tile)
			(let* ((pos (cons x y))
			       (neihbor (hash-get hash pos (lambda () false))))
			  (if neihbor (union p neihbor)
			      (hash-put! hash pos p))))))))

      (for-each (lambda (p) 
		  (let ((x (player-x (union-data p)))
			(y (player-y (union-data p))))
		    (apply p x y)

		    (apply p (+ x 1) y)
		    (apply p (- x 1) y)
		    (apply p x (+ y 1))
		    (apply p x (- y 1))

		    (apply p (+ x 1) (+ y 1))
		    (apply p (- x 1) (- y 1)) 
		    (apply p (- x 1) (+ y 1))
		    (apply p (+ x 1) (- y 1))))
		players)
      
      (union-groups players)))

  (define (my-cluster board players my-id)
    (define (is-my-group? group)
      (ormap (lambda (p) (if (= (player-id p) my-id) group false)) group))

    (ormap is-my-group? (cluster board players)))

  (define (is-occupable-tile tile)
    (or (eq? tile 'plain)
	(eq? tile 'home)))

  (define (player-possible-moves board p)
    (let loop ((dirs '(N S E W)) (rtn empty))
      (if (empty? dirs) (cons 'none rtn)
	  (loop (rest dirs) 
		(if (is-occupable-tile
		     (board-ref board
				(next-x (player-x p) (first dirs))
				(next-y (player-y p) (first dirs))))
		    (cons (first dirs) rtn)
		    rtn)))))
  
  (define (test)
    (define b (read-board-from-file "test.map"))
    (define pls (list (make-player 1 2 1 empty)
		      (make-player 2 5 5 empty)
		      (make-player 3 2 3 empty)
		      (make-player 4 1 7 empty)
		      ))

    (show-board b pls) (newline) (newline) (newline) 
    (debug-f "->" (cluster b pls))
    (debug-f "->" (my-cluster b pls 3)))

;  (test)     

)
