; vim: ft=scheme

(module dijkstra mzscheme

  (require (lib "list.ss")
           (lib "posn.ss" "lang")
           (lib "match.ss")
           (file "../twooldri/data.ss")
           (file "../gmarceau/heap.ss")
           (file "../rob/base.ss")
           (file "../dbt/board.ss"))

  (provide (all-defined))

  (define *infinity* 10e7)

  (define *water-weight* 20)
  (define *wall-weight* 10)

  (define *the-priority-queue* 
    (make-heap (lambda (p1 p2) (< (get-weight-from-source p1) (get-weight-from-source p2)))
               (lambda (p1 p2) (and (eq? (posn-x p1) (posn-x p2)) (eq? (posn-y p1) (posn-y p2))))))

  (define (is-on-board? board posn) 
    (let ((max-x (board-width board))
          (max-y (board-height board)))
      (and (<= (posn-x posn) max-x)
           (<= (posn-y posn) max-y)
           (>= (posn-x posn) 1)
           (>= (posn-y posn) 1))))

  (define (can-move-to? board posn)
    (let ((pos-type (board-ref board (posn-x posn) (posn-y posn))))
      (and (is-on-board? board posn)
           (not (eq? pos-type 'wall))
           (not (eq? pos-type 'water)))))

  (define (relax posn-u posn-v weight)
    (let ((dist-u (get-weight-from-source posn-u))
          (dist-v (get-weight-from-source posn-v)))

      (when (< (+ dist-u weight) dist-v)
        (set-weight-from-source posn-v (+ dist-u weight))
        (set-predecessor-of posn-v posn-u)
        (heap-resort *the-priority-queue* posn-v))))

  (define (initialize-single-source board source-posn)

    (let loop ((x 1) (y 1))

      (let ((current-posn (make-posn x y)))
        (set-weight-from-source current-posn *infinity*)
        (set-predecessor-of current-posn 'undefined)
        (heap-insert *the-priority-queue* current-posn))

      (unless (and (eq? x (board-width board)) 
                   (eq? y (board-height board)))
        (if (eq? x (board-width board))
          (loop 1 (add1 y))
          (loop (add1 x) y))))

    (set-weight-from-source source-posn 0)
    (set-predecessor-of source-posn 'undefined)
    (heap-resort *the-priority-queue* source-posn))

  (define (dijkstra board source-posn) 

    (initialize-single-source board source-posn)

    (unless (eq? (heap-size *the-priority-queue*) 0)
      (let loop ((posn1 (heap-pop *the-priority-queue*)))

        (match-let ((($ posn x y) posn1))
          (let ((posn-left (make-posn (- x 1) y))
                (posn-right (make-posn (+ x 1) y))
                (posn-up (make-posn x (+ y 1)))
                (posn-down (make-posn x (- y 1))))

            ;; boundary checking..
            (when (can-move-to? board posn-left)
              (relax posn1 posn-left (get-weight-between posn1 posn-left)))
            (when (can-move-to? board posn-right)
              (relax posn1 posn-right (get-weight-between posn1 posn-right)))
            (when (can-move-to? board posn-up)
              (relax posn1 posn-up (get-weight-between posn1 posn-up)))
            (when (can-move-to? board posn-down)
              (relax posn1 posn-down (get-weight-between posn1 posn-down)))

            (unless (eq? (heap-size *the-priority-queue*) 0)
              (loop (heap-pop *the-priority-queue*))))))))


  (define (set-all-weights-to-one board)

    (let loop ((current-posn (make-posn 1 1)))
      (match-let ((($ posn currx curry) current-posn))

        (unless (> curry (board-height board))
          (if (> currx (board-width board))
            (loop (make-posn 1 (add1 curry)))
            (begin
              (set-weight-between current-posn (make-posn (add1 currx) curry) 1)
              (set-weight-between current-posn (make-posn currx (add1 curry)) 1)

              (loop (make-posn (add1 currx) curry))))))))


  ;; rewritten for speed.
  (define (propogate-weight board current-posn weight)

    (match-let ((($ posn x y) current-posn))
      (let ((posn-left1 (make-posn (sub1 x) y))
            (posn-left2 (make-posn (- x 2) y))
            (posn-right1 (make-posn (add1 x) y))
            (posn-right2 (make-posn (+ x 2) y))
            (posn-up1 (make-posn x (add1 y)))
            (posn-up2 (make-posn x (+ y 2)))
            (posn-down1 (make-posn x (sub1 y)))
            (posn-down2 (make-posn x (- y 2))))

        ;; we don't want to add weights to things we can't move to.
        (when (and (can-move-to? board posn-left1)
                   (can-move-to? board posn-left2))
          (set-weight-between posn-left1 posn-left2 (+ (get-weight-between posn-left1 posn-left2) weight)))

        (when (and (can-move-to? board posn-right1)
                   (can-move-to? board posn-right2))
          (set-weight-between posn-right1 posn-right2 (+ (get-weight-between posn-right1 posn-right2) weight)))

        (when (and (can-move-to? board posn-up1)
                   (can-move-to? board posn-up2))
          (set-weight-between posn-up1 posn-up2 (+ (get-weight-between posn-up1 posn-up2) weight)))

        (when (and (can-move-to? board posn-down1)
                   (can-move-to? board posn-down2))
          (set-weight-between posn-down1 posn-down2 (+ (get-weight-between posn-down1 posn-down2) weight))))))


  (define (set-weights-using-board board)
    (set-all-weights-to-one board) ;; hackish. to fix.

    (let loop ((x 0) (y 0))

      ;; hope this syntax is right.
      (let ((pos-type (board-ref board x y)))
        (when (eq? pos-type 'wall)
          (propogate-weight board (make-posn x y) *wall-weight*))
        (when (eq? pos-type 'water)
          (propogate-weight board (make-posn x y) *water-weight*)))

      ;;`the edges of the world behave as if there were walls beyond them', so
      ;; we also need to propogate the weights from the walls directly next
      ;; to the board boundaries.

      (unless (and (> (add1 x) (add1 (board-width board)))
                   (> (add1 y) (add1 (board-height board))))
        (if (> (add1 x) (add1 (board-width board)))
          (loop 0 (add1 y))
          (loop (add1 x) y)))))

  (define (posn=? p1 p2)
    (and (eq? (posn-x p1) (posn-x p2))
         (eq? (posn-y p1) (posn-y p2))))

  ;; board -> (board posn -> (posn -> direction))
  (define (get-path-distance-and-direction-functions board)

    (set-all-weights-to-one board)

    ;; board posn -> (values (posn->distance) (posn -> direction))
    (lambda (my-source-their-dest)
      (dijkstra board my-source-their-dest)

      (values 
        ;; posn -> distance
        get-weight-from-source

        ;; posn posn -> direction
        (lambda (current-posn my-dest-their-source)

          (let ((last-pred my-dest-their-source))
            (let loop ((pred (get-predecessor-of my-dest-their-source)))

              (if (not (posn? pred)) ;; there might not have been a valid path..
                'no-move
                (begin
                  (if (posn=? pred current-posn)
                    (begin
                      (match-let ((($ posn current-x current-y) current-posn)
                                  (($ posn new-x new-y) last-pred))
                        (cond ((> new-x current-x) 'east)                            
                              ((< new-x current-x) 'west)                            
                              ((> new-y current-y) 'north)                           
                              ((< new-y current-y) 'south)                           
                              (else 'no-move))))
                    (begin
                      (set! last-pred pred)
                      (loop (get-predecessor-of pred)))))))))))))


